/*
 * pbs_accounting - program to summarize and report system usage based on
 *                  PBS accounting input data.
 *
 * pbs_accounting revision:
 *                - Albino Aveleda, NACAD-COPPE/UFRJ, bino@nacad.ufrj.br
 *                  12 Oct 2003: fix a bug in clean pointer
 *                  28 Oct 2002: sort reports
 *                  25 Oct 2002: original cluster version (based on
 *                               pbs_acct.c from James P Jones, NASA/Ames)
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#define MAX_HOSTNAME_SIZE  256      /* max length of hostname */
#define MAX_NAME_SIZE      20       /* max length of name     */

/* set this to 1 to report usage in cputime rather than walltime */
#define USE_CPUTIME 0

typedef struct group GROUP;
typedef struct user  USER;

struct group {
   char groupname[MAX_NAME_SIZE];
   int num_jobs;
   double cpu_time;
   double all_cpu_time;
   GROUP *p_next;
};

struct user {
   char username[MAX_NAME_SIZE];
   int num_jobs;
   double cpu_time;
   double all_cpu_time;
   GROUP *pg;
   USER *p_next;
};

/* Allocation new group */
GROUP * new_group()
{
   GROUP *ptr;

   ptr = (GROUP *)malloc(sizeof(GROUP));
   if (ptr != NULL) {
     ptr->groupname[0] = '\0';
     ptr->num_jobs = 0;
     ptr->cpu_time = 0.0;
     ptr->all_cpu_time = 0.0;
     ptr->p_next = NULL;
     return(ptr);
   } else
     return(NULL);
}

/* Allocation new user */
USER * new_user()
{
   USER *ptr;

   ptr = (USER *)malloc(sizeof(USER));
   if (ptr != NULL) {
     ptr->username[0] = '\0';
     ptr->num_jobs = 0;
     ptr->cpu_time =0.0;
     ptr->all_cpu_time =0.0;
     ptr->pg = NULL;
     ptr->p_next = NULL;
     return(ptr);
   } else 
     return(NULL);
}

/* sort groups */
void sort_groups(GROUP *ptr)
{
    GROUP *p, *p1, *p2;

    p  = ptr;
    p1 = ptr;
    while (p1->p_next != NULL){
      p2 = p1->p_next;
      while (p2 != NULL){
        if (strcmp(p1->groupname,p2->groupname) > 0){
	   p1->p_next = p2->p_next;
	   p2->p_next = p1;
	   if (ptr == p1){ 
             ptr = p2;
	   } else {
             p->p_next = p2;}
         }
         p2 = p2->p_next;
      }
      p1 = p1->p_next;
    }
}

/* Clean groups */
void clean_groups(GROUP *ptr)
{
    GROUP *ptr_aux;

    while (ptr != NULL) {
       ptr_aux = ptr;
       ptr = ptr->p_next;
       free(ptr_aux);
    }
}

/* Clean users */
void clean_users(USER *ptr)
{
    USER *ptr_aux;

    while (ptr != NULL) {
       ptr_aux = ptr;
       ptr = ptr->p_next;
       free(ptr_aux);
    }
}

/* Main program */
int main (int argc, char **argv)
{
   int    len, nodes, hour, min, sec, test, linenum = 0;
   double cpu_used;
   char   buffer[BUFSIZ], *p, *ptr, *tmpline;
   char   uid[MAX_NAME_SIZE], gid[MAX_NAME_SIZE];
   char   myname[MAX_HOSTNAME_SIZE + 1] = {0};
   GROUP  *pgroups, *pg, *pg_aux;
   USER   *pusers , *pu, *pu_aux;
   FILE   *f_in;
   
/* check if there is argument */
   if (argc != 2){
     printf("\tUsage: pbs_accounting PBS_Data_File\n");
     exit(1);
   }

/* get the hostname */
   gethostname(myname, sizeof(myname));

/* initialize pointers */

   pgroups = NULL;
   pusers  = NULL;

/* open PBS data file */
   if ((f_in = fopen(argv[1],"r")) == NULL) {
     printf("\tCan't open file %s.\n",argv[1]);
     exit(1);
   }

/* read file */
   if(f_in != NULL) {
     while(fgets(buffer, BUFSIZ, f_in) != NULL) {
        cpu_used = 0.0;
        linenum++;
        len = strlen(buffer) + 1;
        tmpline = malloc(len * sizeof(char));
        strcpy(tmpline, buffer);

        ptr = strstr(tmpline, "Resources_List.ncpus");
        if (ptr == NULL) {
            /* hummm. no List.ncpus entry, lets try node-count */
            strcpy(tmpline, buffer);
            ptr = strstr(tmpline, "Resource_List.nodect");
        }
        /* list of nodes run on */
        if (ptr != NULL) {
            p = strtok(ptr, "=");
            p = strtok(NULL, "=");
            nodes = atoi(p);
        } else {
            nodes = 1;
        }
        /* search for user name */
        strcpy(tmpline, buffer);
        ptr = strstr(tmpline, ";user");
        if (ptr == NULL) {
            fprintf(stderr,
                "Warning: No user (id) data on line %d, skipping...\n",
                linenum);
            free(tmpline);
        }
        p = strtok(ptr, "="); 
        p = strtok(NULL, "=");
        sscanf(p, "%s ", uid);
        p = strtok(NULL, "=");
        sscanf(p, "%s ", gid);

        /* search for cpu or wall time */
#if USE_CPUTIME
        strcpy(tmpline, buffer); 
        ptr = strstr(tmpline, "resources_used.cput");
        if (ptr == NULL) {
            fprintf(stderr,
                "Warning: No used.cput data on line %d, skipping...\n",
            linenum);
            free(tmpline);
        }
        p = strtok(ptr, "="); 
        p = strtok(NULL, "=");
        cpu_used = (double)atof(p)/60; /* cpu minutes */

#else  /* use walltime instead */
        strcpy(tmpline, buffer); 
        ptr = strstr(tmpline, "resources_used.walltime");
        if (ptr == NULL) {
            fprintf(stderr,
                "Warning: No used.walltime data on line %d, skipping...\n",
            linenum);
            free(tmpline);
        }
        p = strtok(ptr, "="); 
        p = strtok(NULL, "=");
        sscanf(p, "%d:%d:%d ", &hour, &min, &sec);
        /* walltime minutes */
        cpu_used = (double)((double)sec/60) + min + (hour * 60);
#endif

/* only for debug
 printf("nodes=%d\tgroup=%s\tuser=%s\tcpu=%10.4f\n",nodes,gid,uid,cpu_used); */
        /* lookup group */
        if (pgroups == NULL){
           pgroups = new_group();
	   pg = pgroups;
           strcpy(pg->groupname,gid);
        } else {
           pg = pgroups;
	   test = strcmp(gid,pg->groupname);
	   if (test != 0) {
	      if (test < 0){
	         pgroups = new_group();
	         pgroups->p_next = pg;
	         pg = pgroups;
	         strcpy(pg->groupname,gid);
	      } else {
		 while ((pg->p_next != NULL) && (strcmp(gid,pg->p_next->groupname) >= 0))
	            pg = pg->p_next;
                 if (strcmp(gid, pg->groupname) > 0) {
		    if (pg->p_next == NULL){
		       pg->p_next = new_group();
		    } else {	    
	               pg_aux = new_group();
		       pg_aux->p_next = pg->p_next;
		       pg->p_next = pg_aux;
		    }
	            pg = pg->p_next;
                    strcpy(pg->groupname,gid);
	         }
              }
	   }
	}
        pg->num_jobs++;
        pg->cpu_time += cpu_used;
        pg->all_cpu_time += nodes * cpu_used;

        /* lookup user */
        if (pusers == NULL){
           pusers = new_user();
	   pu = pusers;
           strcpy(pu->username,uid);
        } else {
           pu = pusers;
	   test = strcmp(uid,pu->username);
	   if (test != 0 ) {
              if (test < 0) {
		 pusers = new_user();
	         pusers->p_next = pu;
	         pu = pusers;
	         strcpy(pu->username,uid);
	      } else {
		 while ((pu->p_next != NULL) && (strcmp(uid,pu->p_next->username) >= 0))
	            pu = pu->p_next;
                 if (strcmp(uid, pu->username) > 0) {
		    if (pu->p_next == NULL){
		       pu->p_next = new_user();
		    } else {	    
	               pu_aux = new_user();
		       pu_aux->p_next = pu->p_next;
		       pu->p_next = pu_aux;
		    }
	            pu = pu->p_next;
                    strcpy(pu->username,uid);
	         }
              }
	   }
	}
        pu->num_jobs++;
        pu->cpu_time += cpu_used;
        pu->all_cpu_time += nodes * cpu_used;
        pu->pg = pg;
     }
   }

/* close PBS data file */
   fclose(f_in);

/* print report user */
   printf("\n*** Portable Batch System accounting statistics ***\n");
   printf("Server Name: %s\n", myname);
   printf("\n*** PBS Per-User Usage Report ***\n");
   printf("User       Group     #Jobs     Wall-Hours      Nodes-Hours\n");
   printf("__________________________________________________________\n");
   pu = pusers;
   while (pu != NULL){
      printf("%-10s %-10s %4d %14.4f %16.4f\n",pu->username,pu->pg->groupname,
             pu->num_jobs,(pu->cpu_time/60),(pu->all_cpu_time/60));
      pu = pu->p_next;
   }

/* print report group */

   printf("\n*** PBS Per-Group Usage Report ***\n");
   printf("Group     #Jobs     Wall-Hours      Nodes-Hours\n");
   printf("_______________________________________________\n");
   pg = pgroups;
   while (pg != NULL){
      printf("%-10s %4d %14.4f %16.4f\n",pg->groupname,
             pg->num_jobs,(pg->cpu_time/60),(pg->all_cpu_time/60));
      pg = pg->p_next;
   }

   clean_groups(pgroups);
   clean_users(pusers);
}
